#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "ExprEvaluator.h"
#include "Tokenizer.h"
#include "stack_ve.h"
#include "stack_ve_sym.h"

OperandType getOperandFromToken(Token t);
SymbolType getSymbolFromToken(Token t);
boolean getOperands(OperandStack evalStack, OperandType *operand1, OperandType *operand2);
OperandType evaluate(OperandType operand1, OperandType operand2, char op);

boolean isOperator(SymbolType symbol);
int getOperatorPriority(SymbolType op);
int compareOpPriority(SymbolType op1, SymbolType op2);

void appendNumber(char expr[], Token t);
void appendSymbol(char expr[], SymbolType symbol);

boolean postfixEval(char expr[], OperandType *value)
{
    boolean result;
    OperandStack evalStack = newOperandStack();
    Tokenizer *tok = newTokenizer(expr);
    Token t;
    while (readNextToken(tok))
    {
        switch (getCurrentTokenType(tok))
        {
        case Number:
            {
                OperandType number;
                getCurrentToken(tok, t);
                number = getOperandFromToken(t);
                operandPush(number, evalStack);
                break;
            }
        case Symbol:
            {
                OperandType operand1, operand2;
                SymbolType op;
                getCurrentToken(tok, t);
                op = getSymbolFromToken(t);
                if (!getOperands(evalStack, &operand1, &operand2))
                {
                    destroyOperandStack(evalStack);
                    destroyTokenizer(tok);
                    return false;          
                }
                operandPush(evaluate(operand1, operand2, op), evalStack);
                break;
            }
        default:
            assert(false);
        }
    }
    if (!isEmptyOperandStack(evalStack))
    {
        *value = operandPop(evalStack);
        result = isEmptyOperandStack(evalStack);
    }
    else
        result = false;
    destroyOperandStack(evalStack);
    destroyTokenizer(tok);
    return result;
}

boolean getOperands(OperandStack evalStack, OperandType *operand1, OperandType *operand2)
{
    if (isEmptyOperandStack(evalStack))
        return false;
    *operand2 = operandPop(evalStack);
    if (isEmptyOperandStack(evalStack))
        return false;
    *operand1 = operandPop(evalStack);
    return true;
}

void appendNumber(char expr[], Token t)
{
    strcat(expr, t);
    strcat(expr, " ");
}

void appendSymbol(char expr[], SymbolType symbol)
{
    int len = strlen(expr);
    expr[len] = symbol;
    expr[len + 1] = ' ';
    expr[len + 2] = '\0';
}

boolean infixToPostfix(char inExpr[], char outExpr[])
{
    boolean result;
    SymbolStack symStack = newSymbolStack();
    Tokenizer *tok = newTokenizer(inExpr);
    Token t;
    outExpr[0] = '\0';

    while (readNextToken(tok))
    {
        switch (getCurrentTokenType(tok))
        {
        case Number:
            {
                getCurrentToken(tok, t);
                appendNumber(outExpr, t);
                break;
            }
        case Symbol:
            {
                SymbolType symbol;
                getCurrentToken(tok, t);
                symbol = getSymbolFromToken(t);
                if (isOperator(symbol))
                {
                    while (!isEmptySymbolStack(symStack) && isOperator(symbolPeek(symStack)) && compareOpPriority(symbol, symbolPeek(symStack)) <= 0)
                    {
                        SymbolType firstSymbolInStack = symbolPop(symStack);    //Estrae l'operatore gi letto con peek
                        appendSymbol(outExpr, firstSymbolInStack);
                    }
                    symbolPush(symbol, symStack);
                }
                else
                {
                    switch (symbol)
                    {
                    case '(':
                        {
                            symbolPush(symbol, symStack);
                            break;
                        }
                    case ')':
                        {
                            SymbolType firstSymbolInStack;
                            for (firstSymbolInStack = symbolPop(symStack); firstSymbolInStack != '('; firstSymbolInStack = symbolPop(symStack))
                            {
                                appendSymbol(outExpr, firstSymbolInStack);
                            }
                            break;
                        }
                    }
                }
                break;
            }
        }
    }
    while (!isEmptySymbolStack(symStack))
    {
        SymbolType op = symbolPop(symStack);
        appendSymbol(outExpr, op);
    }
    result = isEmptySymbolStack(symStack);
    destroySymbolStack(symStack);
    destroyTokenizer(tok);
    return result;
}

boolean infixEval(char expr[], int *value)
{
    boolean result;
    OperandStack evalStack = newOperandStack();
    SymbolStack symStack = newSymbolStack();
    Tokenizer *tok = newTokenizer(expr);
    Token t;

    while (readNextToken(tok))
    {
        switch (getCurrentTokenType(tok))
        {
        case Number:
            {
                OperandType number;
                getCurrentToken(tok, t);
                number = getOperandFromToken(t);
                operandPush(number, evalStack);
                break;
            }
        case Symbol:
            {
                SymbolType symbol;
                getCurrentToken(tok, t);
                symbol = getSymbolFromToken(t);
                if (isOperator(symbol))
                {
                    while (!isEmptySymbolStack(symStack) && isOperator(symbolPeek(symStack)) && compareOpPriority(symbol, symbolPeek(symStack)) <= 0)
                    {
                        OperandType operand1, operand2;
                        SymbolType firstSymbolInStack = symbolPop(symStack);    //Estrae l'operatore gi letto con peek
                        if (!getOperands(evalStack, &operand1, &operand2))
                        {
                            destroyOperandStack(evalStack);
                            destroySymbolStack(symStack);
                            destroyTokenizer(tok);
                            return false;
                        }
                        operandPush(evaluate(operand1, operand2, firstSymbolInStack), evalStack);
                    }
                    symbolPush(symbol, symStack);
                }
                else
                {
                    switch (symbol)
                    {
                    case '(':
                        {
                            symbolPush(symbol, symStack);
                            break;
                        }
                    case ')':
                        {
                            SymbolType firstSymbolInStack;
                            for (firstSymbolInStack = symbolPop(symStack); firstSymbolInStack != '('; firstSymbolInStack = symbolPop(symStack))
                            {
                                OperandType operand1, operand2;
                                if (!getOperands(evalStack, &operand1, &operand2))
                                {
                                    destroyOperandStack(evalStack);
                                    destroySymbolStack(symStack);
                                    destroyTokenizer(tok);
                                    return false;
                                }
                                operandPush(evaluate(operand1, operand2, firstSymbolInStack), evalStack);
                            }
                            break;
                        }
                    }
                }
                break;
            }
        }
    }
    while (!isEmptySymbolStack(symStack))
    {
        SymbolType op = symbolPop(symStack);
        OperandType operand1, operand2;
        if (!getOperands(evalStack, &operand1, &operand2))
        {
            destroyOperandStack(evalStack);
            destroySymbolStack(symStack);
            destroyTokenizer(tok);
            return false;
        }
        operandPush(evaluate(operand1, operand2, op), evalStack);
    }
    if (!isEmptyOperandStack(evalStack))
        *value = operandPop(evalStack);
    result = isEmptySymbolStack(symStack) && isEmptyOperandStack(evalStack);
    destroyOperandStack(evalStack);
    destroySymbolStack(symStack);
    destroyTokenizer(tok);
    return result;
}

OperandType getOperandFromToken(Token t)
{
    OperandType operand;
    sscanf(t, OPERAND_FORMAT, &operand);
    return operand;
}

SymbolType getSymbolFromToken(Token t)
{
    return t[0];
}

OperandType evaluate(OperandType operand1, OperandType operand2, char op)
{
    switch (op)
    {
    case '+':
        return operand1 + operand2;
    case '-':
        return  operand1 - operand2;
    case '*':
        return operand1 * operand2;
    case '/':
        return operand1 / operand2;
    }
    abort();
    return 0;
}

boolean isOperator(SymbolType symbol)
{
    return symbol == '+' || symbol == '*' || symbol == '-' || symbol == '/';
}

int getOperatorPriority(SymbolType op)
{
    switch (op)
    {
    case '+':
    case '-':
        return 1;
    case '*':
    case '/':
        return 2;
    default:
        abort();
        return 0;
    }
}

int compareOpPriority(SymbolType op1, SymbolType op2)
{
    return getOperatorPriority(op1) - getOperatorPriority(op2);
}

void postfixEvalTest()
{
    int value;
    assert(postfixEval("10 2 + 1 4 * /", &value));
    assert(value == 3);
    assert(!postfixEval("10 2 4 +", &value));
}

void infixToPostfixTest()
{
    char outExpr[30];
    assert(infixToPostfix("10 - 2 * 3", outExpr));
    assert(strcmp("10 2 3 * - ", outExpr) == 0);

    assert(infixToPostfix("10 * 2 - 3", outExpr));
    assert(strcmp("10 2 * 3 - ", outExpr) == 0);

    assert(infixToPostfix("2 * (3 + 4)", outExpr));
    assert(strcmp("2 3 4 + * ", outExpr) == 0);

    assert(infixToPostfix("10 - 2 - 3 * 2 / (2 * 1)", outExpr));
    assert(strcmp("10 2 - 3 2 * 2 1 * / - ", outExpr) == 0);
}

void infixEvalTest()
{
    int value;

    assert(infixEval("10 - 2 * 3", &value));
    assert(value == 4);

    assert(infixEval("10 * 2 - 3", &value));
    assert(value == 17);

    assert(infixEval("2 * (3 + 4)", &value));
    assert(value == 14);

    assert(infixEval("10 - 2 - 3 * 2 / (2 * 1)", &value));
    assert(value == 5);
}