#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "Tokenizer.h"

boolean readNumber(char *buffer, int *readChars, Token number);

Tokenizer* newTokenizer(char expression[EXP_BUFFER_DIM])
{
    Tokenizer *tokenizer = (Tokenizer*)malloc(sizeof(Tokenizer));
    strcpy(tokenizer->buffer, expression);
    tokenizer->currentIdx = 0;
    strcpy(tokenizer->currentToken, "");
    tokenizer->currentTokenType = None;
    return tokenizer;
}

void destroyTokenizer(Tokenizer *tokenizer)
{
    free(tokenizer);
}

boolean readNextToken(Tokenizer* tokenizer)
{
    int readChars;
    Token t;
    while (tokenizer->buffer[tokenizer->currentIdx] != '\0' &&
        tokenizer->buffer[tokenizer->currentIdx] == ' ')    //trimLeft
    {
        tokenizer->currentIdx++;
    }
    if (tokenizer->buffer[tokenizer->currentIdx] == '\0')
        return false;
    if (readNumber(tokenizer->buffer + tokenizer->currentIdx, &readChars, t))
    {
        strcpy(tokenizer->currentToken, t);
        tokenizer->currentTokenType = Number;
        tokenizer->currentIdx += readChars;
    }
    else
    {
        tokenizer->currentToken[0] = tokenizer->buffer[tokenizer->currentIdx];
        tokenizer->currentToken[1] = '\0';
        tokenizer->currentTokenType = Symbol;
        tokenizer->currentIdx++;
    }
    return true;
}

void getCurrentToken(Tokenizer* tokenizer, Token t)
{
    strcpy(t, tokenizer->currentToken);
}

TokenType getCurrentTokenType(Tokenizer* tokenizer)
{
    return tokenizer->currentTokenType;
}

boolean readNumber(char *buffer, int *readChars, Token number)
{
   int i;
   for (i = 0; buffer[i] >= '0' && buffer[i] <= '9'; i++)
       number[i] = buffer[i];
   if (i > 0)
   {
       number[i] = '\0';
       *readChars = i;
       return true;
   }
   return false;
}

void tokenizerTest()
{
    Token t;
    Tokenizer *tizer = newTokenizer("10 + 3   / 1234 (  ");
    
    getCurrentToken(tizer, t);
    assert(t[0] == '\0');

    assert(readNextToken(tizer));
    assert(getCurrentTokenType(tizer) == Number);
    getCurrentToken(tizer, t);
    assert(strcmp(t, "10") == 0);

    assert(readNextToken(tizer));
    assert(getCurrentTokenType(tizer) == Symbol);
    getCurrentToken(tizer, t);
    assert(strcmp(t, "+") == 0);

    assert(readNextToken(tizer));
    assert(getCurrentTokenType(tizer) == Number);
    getCurrentToken(tizer, t);
    assert(strcmp(t, "3") == 0);

    assert(readNextToken(tizer));
    assert(getCurrentTokenType(tizer) == Symbol);
    getCurrentToken(tizer, t);
    assert(strcmp(t, "/") == 0);

    assert(readNextToken(tizer));
    assert(getCurrentTokenType(tizer) == Number);
    getCurrentToken(tizer, t);
    assert(strcmp(t, "1234") == 0);

    assert(readNextToken(tizer));
    assert(getCurrentTokenType(tizer) == Symbol);
    getCurrentToken(tizer, t);
    assert(strcmp(t, "(") == 0);

    assert(!readNextToken(tizer));
}

